import sys
import pandas as pd
from typing import List, Any, Optional
from bioforklift.forklift_logging import setup_logger
from .terra_entities import TerraEntities

logger = setup_logger(__name__)


class TerraMerge:
    """
    Class to handle merging of Terra datatables.
    This class is used to identify samples present in multiple tables,
    merge their attributes, and prepare the result for upload to a master table.
    """

    def __init__(self, terra_entities_client: TerraEntities):
        """
        Initialize the TerraMerge class.

        Args:
            terra_client: An initialized Terra client
        """
        self.terra_entities = terra_entities_client

    def merge_and_update_master(
        self,
        primary_table: str,
        secondary_table: str,
        master_table: str,
        primary_key: Optional[str] = None,
        secondary_key: Optional[str] = None,
        master_key: Optional[str] = None,
        fields_to_include: Optional[List[str]] = None,
        fields_to_exclude: Optional[List[str]] = None,
        use_destination: bool = False,
    ) -> pd.DataFrame:
        """
        Merge two tables and update a master table with new records.

        Args:
            primary_table: Primary table name (e.g., "sample")
            secondary_table: Secondary table name (e.g., "sample_metadata")
            master_table: Master table to upload merged results to (e.g., "sample_master")
            primary_key: Key field in primary table (auto-detected if None)
            secondary_key: Key field in secondary table (auto-detected if None)
            master_key: Key field for upload to master table (auto-detected if None)
            fields_to_include: List of specific fields to include in merge
            fields_to_exclude: List of fields to exclude from merge
            use_destination: Whether to use destination workspace

        Returns:
            DataFrame with newly added records
        """
        logger.info(
            f"Starting merge of {primary_table} and {secondary_table} for update to {master_table}"
        )

        # Download the primary and secondary tables , usually will be something like a sample table
        # With reads and a metadata table
        primary_df = self.terra_entities.download_table(
            primary_table, use_destination=use_destination
        )
        logger.info(
            f"Downloaded primary table: {primary_table} with {len(primary_df)} records"
        )

        secondary_df = self.terra_entities.download_table(
            secondary_table, use_destination=use_destination
        )
        logger.info(
            f"Downloaded secondary table: {secondary_table} with {len(secondary_df)} records"
        )

        try:
            master_df = self.terra_entities.download_table(
                master_table, use_destination=use_destination
            )
            logger.info(
                f"Downloaded master table: {master_table} with {len(master_df)} records"
            )
        except Exception as exc:
            # If master table doesn't exist or is empty, create an empty DataFrame, e.g., for new uploads
            logger.info(
                f"Master table {master_table} not found or empty, will create new"
            )
            master_df = pd.DataFrame()

        # Auto-enforce keys if not provided, we can do this automatically from the entity ids
        if primary_key is None:
            primary_key = f"entity:{primary_table}_id"
            logger.info(f"Using detected primary key: {primary_key}")

        if secondary_key is None:
            secondary_key = f"entity:{secondary_table}_id"
            logger.info(f"Using detected secondary key: {secondary_key}")

        if master_key is None:
            # For master table, we need to ensure the key follows Terra conventions
            master_key = f"entity:{master_table}_id"
            if not master_df.empty and master_key not in master_df.columns:
                # If master table exists but doesn't have expected key, try to detect it
                sys.exit(
                    f"Master table {master_table} does not contain expected key {master_key}"
                )

        # Merge primary and secondary tables where both ids are present
        logger.info(f"Merging tables on {primary_key}={secondary_key}")

        # Get the set of ids from both tables
        primary_ids = set(primary_df[primary_key].dropna())
        secondary_ids = set(secondary_df[secondary_key].dropna())

        # Now we need to find the intersection of these two sets
        common_ids = primary_ids.intersection(secondary_ids)
        logger.info(f"Found {len(common_ids)} sample IDs present in both tables")

        if not common_ids:
            logger.warning(
                "No matching records found between primary and secondary tables"
            )
            return pd.DataFrame()

        # We need to make sure we only keep the records in both tables that match
        primary_filtered = primary_df[primary_df[primary_key].isin(common_ids)]
        secondary_filtered = secondary_df[secondary_df[secondary_key].isin(common_ids)]

        # Prepare for merge by renaming the secondary key to match primary
        if primary_key != secondary_key:
            secondary_filtered = secondary_filtered.rename(
                columns={secondary_key: primary_key}
            )

        # Determine fields to include/exclude
        all_fields = list(
            set(primary_filtered.columns).union(set(secondary_filtered.columns))
        )

        if fields_to_include:
            # If specific fields are included, use only those that exist
            fields_to_use = [
                field for field in fields_to_include if field in all_fields
            ]
            # Always include the primary key
            if primary_key not in fields_to_use:
                fields_to_use.append(primary_key)

            # Filter primary and secondary dataframes to only include specified fields
            primary_fields = [
                field for field in fields_to_use if field in primary_filtered.columns
            ]
            secondary_fields = [
                field for field in fields_to_use if field in secondary_filtered.columns
            ]

            # Ensure primary key is included in both dataframes for merge
            if primary_key not in primary_fields:
                primary_fields.append(primary_key)
            if primary_key not in secondary_fields:
                secondary_fields.append(primary_key)

            primary_filtered = primary_filtered[primary_fields]
            secondary_filtered = secondary_filtered[secondary_fields]
        else:
            # Use all fields except excluded
            fields_to_use = all_fields
            if fields_to_exclude:
                # Apply exclusions , filter dataframes to exclude specified fields
                primary_fields = [
                    field
                    for field in primary_filtered.columns
                    if field not in fields_to_exclude or field == primary_key
                ]
                secondary_fields = [
                    field
                    for field in secondary_filtered.columns
                    if field not in fields_to_exclude or field == primary_key
                ]

                primary_filtered = primary_filtered[primary_fields]
                secondary_filtered = secondary_filtered[secondary_fields]

        # Perform merge - use outer join to ensure all records from both tables are included
        # Only keep records that appear in both tables
        merged_df = pd.merge(
            primary_filtered, secondary_filtered, on=primary_key, how="inner"
        )

        logger.info(
            f"Merged result contains {len(merged_df)} rows and {len(merged_df.columns)} columns"
        )

        if not master_df.empty:
            # Get existing ids from master table
            if master_key in master_df.columns:
                master_ids = set(master_df[master_key].dropna())
            else:
                master_ids = set()

            # Find ids in merged data not already in master
            new_ids = common_ids - master_ids

            if not new_ids:
                logger.info("No new records to add to master table")
                return pd.DataFrame()  # Return empty DataFrame

            logger.info(f"Found {len(new_ids)} new records to add to master table")

            # Filter to only new records
            merged_df = merged_df[merged_df[primary_key].isin(new_ids)]

        # Now we need to ensure the master table has the correct key (entity id)
        if primary_key != master_key:
            merged_df = merged_df.rename(columns={primary_key: master_key})

        # And finally now we can upload the merged data to the master table
        if not merged_df.empty:
            logger.info(f"Uploading {len(merged_df)} new records to {master_table}")
            uploaded_df = self.upload_to_master(
                merged_df=merged_df,
                master_table=master_table,
                id_column=master_key,
                use_destination=use_destination,
            )
            return uploaded_df
        else:
            logger.info("No records to upload after filtering")
            return pd.DataFrame()

    def upload_to_master(
        self,
        merged_df: pd.DataFrame,
        master_table: str,
        id_column: str,
        use_destination: bool = False,
    ) -> pd.DataFrame:
        """
        Upload merged data to master table.

        Args:
            merged_df: DataFrame with merged data
            master_table: Name of master table
            id_column: Column to use as entity ID
            use_destination: Whether to use destination workspace

        Returns:
            DataFrame with upload results
        """
        upload_df = merged_df.copy()

        return self.terra_entities.upload_entities(
            data=upload_df,
            target=master_table,
            entity_identifier_column=id_column,
            use_destination=use_destination,
        )



from typing import Callable, Optional, Set
import pandas as pd
from bioforklift.forklift_logging import setup_logger
from bioforklift.terra.client import TerraClient
from bioforklift.terra.terra_entities import TerraEntities
from bioforklift.terra.models import TransferResult, TransferStatus
from bioforklift.terra.exceptions import (
    TerraTransferSourceError,
    TerraTransferUploadError,
    TerraNotFoundError,
)

logger = setup_logger(__name__)


class TerraToTerraTransfer:
    """Transfer entities between Terra workspaces with deduplication."""

    def __init__(
        self,
        client: TerraClient,
        source_table_name: str,
        destination_table_name: str,
        source_identifier_column: Optional[str] = None,
        destination_identifier_column: Optional[str] = None,
        batch_size: int = 500,
        transform: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None,
    ):
        """
        Initialize TerraToTerraTransfer.

        Args:
            client: TerraClient configured with source and destination workspaces
            source_table_name: Name of the entity table in the source workspace
            destination_table_name: Name of the entity table in the destination workspace
            source_identifier_column: Column used to identify samples in source table.
                Defaults to {source_table_name}_id (Terra convention).
            destination_identifier_column: Column used to identify samples in destination table.
                Defaults to {destination_table_name}_id (Terra convention).
            batch_size: Number of entities per upload batch (default 500)
            transform: Optional callable to transform the DataFrame before upload.
                Receives the filtered DataFrame (new samples only) and should return
                a transformed DataFrame. Use this to filter rows, drop columns,
                or modify values before upload.
        """
        self.client = client
        self.entities = TerraEntities(client)
        self.source_table_name = source_table_name
        self.destination_table_name = destination_table_name
        self.source_identifier_column = source_identifier_column or f"entity:{source_table_name}_id"
        self.destination_identifier_column = destination_identifier_column or f"entity:{destination_table_name}_id"
        self.batch_size = batch_size
        self.transform = transform

    def get_new_sample_ids(self) -> Set[str]:
        """
        Compare source and destination tables to find new sample IDs.

        Returns:
            Set of sample IDs that exist in source but not in destination

        Raises:
            TerraTransferSourceError: If source table cannot be accessed or identifier column is missing
        """

        # Download source table (only identifier column needed for comparison)
        logger.info(f"Fetching sample IDs from source workspace")
        try:
            source_df = self.entities.download_table(
                entity_type=self.source_table_name,
                attributes=[self.source_identifier_column],
                use_destination=False
            )
        except TerraNotFoundError as e:
            raise TerraTransferSourceError(
                f"Source table '{self.source_table_name}' not found in workspace "
                f"'{self.client.source_project}/{self.client.source_workspace}'"
            ) from e
        except Exception as e:
            raise TerraTransferSourceError(
                f"Failed to access source table '{self.source_table_name}': {e}"
            ) from e

        if self.source_identifier_column not in source_df.columns:
            raise TerraTransferSourceError(
                f"Identifier column '{self.source_identifier_column}' not found in source table. "
                f"Available columns: {list(source_df.columns)}"
            )

        # Download destination table (may not exist yet)
        logger.info(f"Fetching sample IDs from destination workspace")
        try:
            dest_df = self.entities.download_table(
                entity_type=self.destination_table_name,
                attributes=[self.destination_identifier_column],
                use_destination=True
            )
        except TerraNotFoundError:
            logger.info(f"Destination table '{self.destination_table_name}' not found, treating as empty")
            dest_df = pd.DataFrame(columns=[self.destination_identifier_column])
        except Exception as e:
            logger.warning(f"Could not access destination table, treating as empty: {e}")
            dest_df = pd.DataFrame(columns=[self.destination_identifier_column])

        # Extract IDs
        source_ids = set(source_df[self.source_identifier_column].dropna().astype(str).tolist())
        dest_ids = set(dest_df[self.destination_identifier_column].dropna().astype(str).tolist()) if self.destination_identifier_column in dest_df.columns else set()
        new_ids = source_ids - dest_ids

        logger.info(f"Found {len(source_ids)} samples in source, {len(dest_ids)} in destination, {len(new_ids)} new")

        return new_ids

    def transfer(self) -> TransferResult:
        """
        Execute the transfer from source to destination workspace.

        Returns:
            TransferResult with status and details of the transfer

        Raises:
            TerraTransferSourceError: If source table cannot be accessed
            TerraTransferUploadError: If upload to destination fails
        """
        # Find samples to transfer
        new_ids = self.get_new_sample_ids()

        if not new_ids:
            logger.info("No new samples to transfer")
            return TransferResult(
                status=TransferStatus.NO_NEW_SAMPLES,
                message="No new samples to transfer"
            )

        logger.info(f"Transferring {len(new_ids)} new samples")

        # Download full source data (all columns)
        try:
            source_df = self.entities.download_table(
                entity_type=self.source_table_name,
                use_destination=False
            )
        except Exception as e:
            raise TerraTransferSourceError(
                f"Failed to download full source data: {e}"
            ) from e

        # Filter to only new samples
        samples_to_transfer = source_df[source_df[self.source_identifier_column].isin(new_ids)].copy()

        logger.info(f"Filtered to {len(samples_to_transfer)} samples for transfer")

        # Apply optional transform
        if self.transform is not None:
            samples_to_transfer = self.transform(samples_to_transfer)
            logger.info(f"After transform: {len(samples_to_transfer)} samples remaining")

            if samples_to_transfer.empty:
                logger.info("No samples remaining after transform")
                return TransferResult(
                    status=TransferStatus.NO_NEW_SAMPLES,
                    message="No samples remaining after transform"
                )

        # Rename source ID column to destination ID column for upload
        samples_to_transfer = samples_to_transfer.rename(columns={self.source_identifier_column: self.destination_identifier_column})

        # Upload to destination
        try:
            self.entities.upload_entities(
                data=samples_to_transfer,
                target=self.destination_table_name,
                entity_identifier_column=self.destination_identifier_column,
                use_destination=True
            )
        except Exception as e:
            logger.error(f"Failed to upload samples to destination: {e}")
            raise TerraTransferUploadError(
                f"Failed to upload {len(samples_to_transfer)} samples to destination "
                f"'{self.client.destination_project}/{self.client.destination_workspace}': {e}"
            ) from e

        transferred_ids = samples_to_transfer[self.destination_identifier_column].tolist()
        logger.info(f"Successfully transferred {len(transferred_ids)} samples")

        return TransferResult(
            status=TransferStatus.SUCCESS,
            transferred_ids=transferred_ids,
            message=f"Successfully transferred {len(transferred_ids)} samples"
        )



import fnmatch
import logging
import argparse
from pathlib import Path
from bioforklift.scripts.configure import CLIConfig
from bioforklift.terra import Terra


logger = logging.getLogger(__name__)


def download_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Define command-line arguments for download subcommand"""
    d_parser = parser.add_argument_group("Data Parameters")
    d_parser.add_argument(
        "table", type=str, nargs="+",help="Terra entity table name(s) (space-delimited)"
    )

    ws_parser = parser.add_argument_group("Terra Workspace Parameters")
    ws_parser.add_argument("-ws", "--workspace", type=str, help="Terra workspace name")
    ws_parser.add_argument("-p", "--project", type=str, help="Terra project name")

    run_parser = parser.add_argument_group("Runtime Parameters")
    run_parser.add_argument(
        "-o",
        "--output_path",
        type=str,
        help="Path to save downloaded data; DEFAULT: current directory",
    )
    return parser


def download(args: argparse.Namespace, config: CLIConfig = CLIConfig()) -> None:
    """Download data from Terra workspace"""
    config.update(vars(args))

    terra = Terra(
        source_workspace=config.workspace,
        source_project=config.project,
        destination_workspace=config.workspace,
        destination_project=config.project,
    )

    # Create output directory if it doesn't exist
    if args.output_path:
        output_dir = Path(args.output_path)
        output_dir.mkdir(parents=True, exist_ok=True)
    else:
        output_dir = Path.cwd()

    terra_entities = terra.entities.list_entity_types(include_attributes=False)
    # unpack the matches via glob-like pattern matching
    tables = []
    for table in args.table:
        fn_tables = fnmatch.filter(terra_entities, table)
        # exclude the _set suffix
        if any(t.endswith("_set") for t in fn_tables):
            logger.debug(f"Excluding '_set' tables from matches for pattern '{table}'")
            fn_tables = [t for t in fn_tables if not t.endswith("_set")]
        logger.debug(f"Pattern '{table}' matched tables: {fn_tables}")
        if not fn_tables:
            raise ValueError(f"No tables match '{table}' in the Terra workspace.")
        tables.extend(fn_tables)

    for table in tables:
        df = terra.entities.download_table(table)
        # Save the downloaded table to defined output path or current directory
        output_path = output_dir / f"{table}.tsv"
        df.to_csv(output_path, index=False, sep="\t")
        logger.info(f"Downloaded table '{table}' to {output_path}")


import logging
import argparse
import pandas as pd
from pathlib import Path
from bioforklift.scripts.configure import CLIConfig
from bioforklift.terra import Terra


logger = logging.getLogger(__name__)


def upload_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Command-line arguments for uploading data to Terra workspace"""
    d_parser = parser.add_argument_group("Data Parameters")
    d_parser.add_argument(
        "input",
        type=str,
        nargs="+",
        help="Path(s) to input data file(s) in Terra workspace (space-delimited)",
    )
    d_parser.add_argument(
        "-t", "--table", 
        type=str, 
        nargs="+",
        help="Terra entity table name(s) (space-delimited)"
    )
    d_parser.add_argument(
        "-o", "--overwrite", action="store_true", help="Overwrite existing entities"
    )

    ws_parser = parser.add_argument_group("Terra Workspace Parameters")
    ws_parser.add_argument("-ws", "--workspace", type=str, help="Terra workspace name")
    ws_parser.add_argument("-p", "--project", type=str, help="Terra project name")
    return parser


def upload(args: argparse.Namespace, config: CLIConfig = CLIConfig()) -> None:
    """Upload data to Terra workspace"""

    config.update(vars(args))

    terra = Terra(
        source_workspace=config.workspace,
        source_project=config.project,
        destination_workspace=config.workspace,
        destination_project=config.project,
    )

    terra_entities = terra.entities.list_entity_types(include_attributes=False)

    if args.table:
        if len(args.input) != len(args.table):
            raise ValueError(
                "Number of input files must match number of table names provided."
            )

    # attempt to infer separator
    for index, in_ in enumerate(args.input):
        if in_.endswith(".tsv"):
            sep = "\t"
        else:
            sep = ","
        # determine table name
        if args.table:
            table_name = args.table[index]
        else:
            table_name = Path(in_).stem

        # upload data
        df = pd.read_csv(in_, sep=sep)
        if not args.overwrite and table_name in terra_entities:
            raise ValueError(
                f"Table '{table_name}' already exists in workspace. Use --overwrite to replace it."
            )
        terra.entities.upload_entities(data=df, target=table_name)
        logger.info(
            f"Uploaded data from '{in_}' to table '{table_name}' in workspace '{config.project}/{config.workspace}'"
        )


from typing import Optional, List, Union, Dict, Any
from pydantic import BaseModel, Field, field_validator
import re


class FieldAttributes(BaseModel):
    """
    Base model for field attributes in schema definitions.

    Contains common attributes shared by both sample and config schemas.
    """

    # Field identification and classification
    primary_key: bool = Field(default=False, description="Field is a primary key (auto-generated UUID)")

    # Data processing attributes
    column_mappings: Optional[List[str]] = Field(
        default=None,
        description="Source column names to map to this field"
    )
    use_field_name: bool = Field(
        default=False,
        description="Use the field name as-is without mapping"
    )

    # Validation attributes
    accepted_pattern: Optional[str] = Field(
        default=None,
        description="Regex pattern for field value validation"
    )
    required: bool = Field(
        default=False,
        description="Field is required"
    )

    # System fields
    system_value: bool = Field(
        default=False,
        description="System-generated value (should be dropped before Terra upload)"
    )

    # Display and formatting
    use_as_prefix: bool = Field(
        default=False,
        description="Use this field value as a prefix"
    )
    display_for_alerts: bool = Field(
        default=False,
        description="Display this field in alerts"
    )
    date_format: Optional[str] = Field(
        default=None,
        description="Date format specification for date/string coercion (e.g., 'ISO 8601')"
    )

    @field_validator('accepted_pattern')
    @classmethod
    def validate_pattern(cls, v: Optional[str]) -> Optional[str]:
        """Validate that pattern is a valid regex."""
        if v is not None:
            try:
                re.compile(v)
            except re.error as e:
                raise ValueError(f"Invalid regex pattern: {e}")
        return v

    @field_validator('column_mappings')
    @classmethod
    def normalize_column_mappings(cls, v: Optional[Union[str, List[str]]]) -> Optional[List[str]]:
        """Normalize column_mappings to always be a list."""
        if v is None:
            return None
        if isinstance(v, str):
            return [v]
        return v

    @field_validator('date_format')
    @classmethod
    def validate_date_format(cls, v: Optional[str]) -> Optional[str]:
        """Validate that date_format is a recognized format."""
        if v is None:
            return None

        # Define supported date format identifiers - can be expanded as needed, but these are knowns
        # from our lab partners
        supported_formats = {
            'ISO 8601',
            'ISO8601',
            'RFC 3339',
            'RFC3339',
            'YYYY-MM-DD',
            'MM/DD/YYYY',
            'DD/MM/YYYY',
            'YYYY/MM/DD',
        }

        # Allow either exact match or strftime format strings
        if v in supported_formats or '%' in v:
            return v

        raise ValueError(
            f"Unsupported date format: {v}. "
            f"Supported formats: {', '.join(sorted(supported_formats))} "
            f"or a valid strftime format string (containing '%')"
        )

    def is_identifier_field(self) -> bool:
        """Check if this field is any type of identifier."""
        return self.primary_key

    def is_system_field(self) -> bool:
        """Check if this is a system-managed field."""
        return self.system_value or self.primary_key


class SampleFieldAttributes(FieldAttributes):
    """
    Extended attributes specific to sample fields.

    Adds sample-specific validation and processing attributes.
    """

    # Sample identification
    sample_identifier: bool = Field(
        default=False,
        description="Field uniquely identifies samples"
    )

    # Metadata and synchronization
    metadata: bool = Field(
        default=False,
        description="Field contains metadata"
    )
    sync_field: bool = Field(
        default=False,
        description="Field should be synchronized to Terra"
    )

    # File and sequence attributes
    sequence_file: bool = Field(
        default=False,
        description="Field contains sequence file path"
    )

    # Configuration inheritance
    inherit_from_config: Optional[str] = Field(
        default=None,
        description="Configuration field to inherit value from"
    )
    
    # Configuration identification for samples
    configuration_identifier: bool = Field(
        default=False,
        description="Field uniquely identifies configurations"
    )

    def is_identifier_field(self) -> bool:
        """Check if this field is any type of identifier."""
        return self.primary_key or self.sample_identifier or self.configuration_identifier

    def should_sync(self) -> bool:
        """Check if this field should be synchronized to Terra."""
        return self.sync_field or self.metadata


class ConfigFieldAttributes(FieldAttributes):
    """
    Extended attributes specific to configuration fields.

    Adds config-specific validation and processing attributes.
    """

    terra_method_config: Optional[Union[str, Dict[str, Any]]] = Field(
        default=None,
        description="Field is part of Terra method configuration (JSON string or dict)"
    )
    single_datatable: bool = Field(
        default=False,
        description="Indicates if source and destination datatables are the same (skips upload step)"
    )


class FieldDefinition(BaseModel):
    """
    Complete field definition including schema and custom attributes.
    """

    name: str = Field(description="Field name")
    field_type: str = Field(description="BigQuery field type (STRING, INTEGER, etc.)")
    mode: str = Field(default="NULLABLE", description="Field mode (REQUIRED, NULLABLE, REPEATED)")
    description: Optional[str] = Field(default=None, description="Field description")

    # Custom attributes
    attributes: FieldAttributes = Field(
        default_factory=FieldAttributes,
        description="Custom processing attributes"
    )

    def has_pattern(self) -> bool:
        """Check if field has a validation pattern."""
        return self.attributes.pattern is not None

    def is_required(self) -> bool:
        """Check if field is required."""
        return self.mode == "REQUIRED" or self.attributes.required

    def get_source_columns(self) -> List[str]:
        """Get list of possible source column names."""
        if self.attributes.column_mappings:
            return self.attributes.column_mappings
        if self.attributes.use_field_name:
            return [self.name]
        return []


class SchemaDefinition(BaseModel):
    """
    Complete schema definition with all fields.
    """

    fields: List[FieldDefinition] = Field(description="List of field definitions")

    def get_field(self, name: str) -> Optional[FieldDefinition]:
        """Get field definition by name."""
        for field in self.fields:
            if field.name == name:
                return field
        return None

    def get_identifier_fields(self) -> List[FieldDefinition]:
        """Get all identifier fields."""
        return [
            field for field in self.fields
            if field.attributes.is_identifier_field()
        ]

    def get_pattern_fields(self) -> List[FieldDefinition]:
        """Get all fields with validation patterns."""
        return [
            field for field in self.fields
            if field.has_pattern()
        ]

    def get_sync_fields(self) -> List[FieldDefinition]:
        """Get all fields that should be synchronized."""
        return [
            field for field in self.fields
            if field.attributes.should_sync()
        ]

    def get_system_fields(self) -> List[FieldDefinition]:
        """Get all system-managed fields."""
        return [
            field for field in self.fields
            if field.attributes.is_system_field()
        ]



from typing import Dict, Any
from .models import MethodConfig, MethodRepoMethod
from .client import TerraClient
from bioforklift.forklift_logging import setup_logger

logger = setup_logger(__name__)


class TerraMethods:
    """
    Class meant to handle Terra workflows and method configurations
    See https://github.com/broadinstitute/fiss/blob/b3fa2a0d888610e04f744f9e661fa32e46a5af95/firecloud/api.py#L650
    """


    def __init__(self, client: TerraClient):
        self.client = client

    def generate_method_config(
        self,
        repo_uri: str,
        wf_name: str,
        table_name: str,
        inputs_dict: dict,
        outputs_dict: dict,
        branch: str,
        source_repo: str = "dockstore",
    ) -> MethodConfig:
        """
        Generate a MethodConfig object for a workflow
        
        Args:
            repo_uri: URI of the repository containing the workflow
            wf_name: Name of the workflow
            table_name: Name of the Terra entity table
            inputs_dict: Dictionary of workflow inputs
            outputs_dict: Dictionary of workflow outputs
            branch: Branch of the repository
            source_repo: Source repository type (default: "dockstore")

        Returns:
            MethodConfig object representing the workflow configuration    
        """
        # to be removed when added natively
        method_config = MethodConfig(
            namespace=self.client.destination_project,
            name=wf_name,
            rootEntityType=f"{table_name}_set",
            inputs=inputs_dict,
            outputs=outputs_dict,
            prerequisites={},
            methodRepoMethod=MethodRepoMethod(
                sourceRepo=source_repo,
                methodPath=f"{repo_uri}/{wf_name}",
                methodVersion=branch,
            ),
            methodConfigVersion=0,
            deleted=False,
        )
        return method_config
    

    def get_method_config(
        self,
        config_name: str,
        use_destination: bool = False,
    ) -> Dict[str, Any]:
        """
        Get a workspace method configuration by name. By default pulls from source workspace.

        Args:
            config_name: Name of the method configuration to retrieve

        Returns:
            Dict containing the method configuration details
        """
        client_project = self.client.destination_project if use_destination else self.client.source_project

        logger.info(f"Fetching workspace method configuration: {config_name} from project: {client_project}")
        response = self.client.get(
          f"method_configs/{client_project}/{config_name}",
          use_destination=use_destination,
        )
        return response.json()


    def overwrite_method_config(
        self,
        config: MethodConfig,
        use_destination: bool = True,
    ) -> Dict[str, Any]:
        """
        Add or overwrite a new workspace method configuration. AKA create a new workflow in Terra with inputs/outputs defined.
        By default pulls from destination workspace.

        Args:
            config: MethodConfig containing all configuration details

        Returns:
            Dict containing the created method configuration details
        """
        client_project = self.client.destination_project if use_destination else self.client.source_project

        logger.info(f"Uploading workspace method configuration: {config.name} to project: {client_project}")
        return self.client.put(
            f"method_configs/{client_project}/{config.name}",
            data=config.model_dump(exclude_none=True),
            use_destination=use_destination,
        ).json()


    def method_config_validate(
      self,
      config: MethodConfig,
      use_destination: bool = True,
    ) -> Dict[str, Any]:
        """
        Validate a workspace method configuration. Note that this can only validate existing configurations in the workspace.
        Meaning you must first call `overwrite_method_config` before validating a new configuration.
        Args:
            config: MethodConfig containing all configuration details

        Returns:
            Dict containing the validation results
        """
        client_project = self.client.destination_project if use_destination else self.client.source_project

        logger.info(f"Validating workspace method configuration: {config.name} from project: {client_project}")
        response = self.client.get(
            f"method_configs/{client_project}/{config.name}/validate",
            use_destination=use_destination,
        ).json()

        # Check for invalid inputs/outputs and raise error if found
        validation_error_keys = ["invalidInputs", "invalidOutputs", "missingInputs", "extraInputs"]
        invalid_fields = {
            key: response.get(key)
            for key in validation_error_keys
            if response.get(key)
        }

        if invalid_fields:
            msg = "MethodConfig validation errors found:\n" + "\n".join(
                f"- {key}: {val}" for key, val in invalid_fields.items()
            )
            logger.error(msg)
            raise ValueError(msg)

        return response


    def dict_to_method_config(
        self,
        config_dict: Dict[str, Any]
    ) -> MethodConfig:
        """
        Convert a dictionary to a MethodConfig object without making an API call.
        Useful for modifying existing configurations.

        Args:
            config_dict: Dict containing method configuration details

        Returns:
            MethodConfig object constructed from the dictionary
        """
        return MethodConfig.model_validate(config_dict)


import json
import time
import logging
import argparse
from pathlib import Path
from datetime import datetime
from bioforklift.terra.exceptions import TerraServerError
from bioforklift.scripts.configure import CLIConfig
from bioforklift.terra import Terra, WorkflowConfig


logger = logging.getLogger(__name__)


def launch_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Define command-line arguments for launch subcommand"""
    wf_parser = parser.add_argument_group("Workflow Submission Parameters")
    wf_parser.add_argument(
        "-wf", "--workflow_name", type=str, help="Terra workflow name to run"
    )
    wf_parser.add_argument(
        "-t", "--table", type=str, help="Terra entity table name for workflow"
    )
    wf_parser.add_argument(
        "-r",
        "--repository",
        type=str,
        help="GitHub repository for workflow source; DEFAULT: github.com/theiagen/public_health_bioinformatics",
    )
    wf_parser.add_argument(
        "-b",
        "--branch",
        type=str,
        help="GitHub branch for workflow source; DEFAULT: main",
    )

    launch_parser = parser.add_argument_group("Workflow Launch Parameters")
    launch_parser.add_argument(
        "-j", "--job_json", type=str, help="Path to workflow submission JSON file"
    )
    launch_parser.add_argument(
        "-s",
        "--sample_col",
        type=str,
        default="samplename",
        help="Column name in entity table for sample names; DEFAULT: samplename",
    )
    launch_parser.add_argument(
        "-i",
        "--input_json",
        type=str,
        help="Path to input JSON file for submission inputs",
    )
    launch_parser.add_argument(
        "-o",
        "--output_json",
        type=str,
        help="Path to output JSON file for submission outputs",
    )
    launch_parser.add_argument(
        "-cc",
        "--call_cache",
        action="store_true",
        help="Enable call caching for the workflow submission",
    )
    launch_parser.add_argument(
        "-c",
        "--comment",
        type=str,
        default="",
        help="User comment for the workflow submission",
    )
    launch_parser.add_argument(
        "-ie",
        "--ignore_empty",
        action="store_true",
        help="Ignore empty outputs in the workflow submission",
    )
    launch_parser.add_argument(
        "--sleep",
        type=int,
        default=1,
        help="Seconds to wait between launching multiple workflows; DEFAULT: 1",
    )

    ws_parser = parser.add_argument_group("Terra Workspace Parameters")
    ws_parser.add_argument("-ws", "--workspace", type=str, help="Terra workspace name")
    ws_parser.add_argument("-p", "--project", type=str, help="Terra project name")
    ws_parser.add_argument(
        "--preexisting_config",
        action="store_true",
        default=False,
        help="Use pre-existing method configuration in the workspace; DEFAULT: False",
    )

    run_parser = parser.add_argument_group("Runtime Parameters")
    run_parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose logging"
    )

    return parser


def prepare_job_dict(args_dict: dict, config: CLIConfig) -> dict:
    """Extract workflow submission parameters from JSON file if provided

    Preference is given to command-line arguments over JSON file values."""

    # argument to requirement
    wf_args = {
        "workflow_name": True,
        "table": True,
        "input_json": True,
        "output_json": True,
        "sample_col": True,
        "comment": False,
        "branch": True,
        "call_cache": True,
        "ignore_empty": True,
        "preexisting_config": True,
    }

    job_dict = {}
    # parse JSON input
    if args_dict.get("job_json"):
        with open(args_dict["job_json"], "r") as json_file:
            json_data = json.load(json_file)
        # iterate through workflows in json file
        for wf, wf_data in json_data.items():
            job_dict[wf] = wf_data
            for arg, require in wf_args.items():
                # use the argument preferentially from command-line
                if args_dict.get(arg) is not None:
                    job_dict[wf][arg] = args_dict.get(arg)
                # else use from json if present
                elif wf_data.get(arg) is not None:
                    job_dict[wf][arg] = wf_data.get(arg)
                # else pull from config if required
                elif require:
                    if config.__dict__.get(arg) is not None:
                        job_dict[wf][arg] = config.__dict__.get(arg)
                    # raise error if still missing
                    else:
                        raise KeyError(
                            f"Missing required argument '{arg}' for workflow '{wf}'"
                        )
    else:
        # single workflow from command-line args
        wf = args_dict["workflow_name"]
        job_dict[wf] = {"workflow_name": wf}
        for arg, require in wf_args.items():
            # use from command-line args if present
            if args_dict.get(arg) is not None:
                if arg == "input_json" or arg == "output_json":
                    with open(args_dict.get(arg), "r") as io_file:
                        job_dict[wf][arg] = json.load(io_file)
                else:
                    job_dict[wf][arg] = args_dict.get(arg)
            # else pull from config if required
            elif require:
                if config.__dict__.get(arg) is not None:
                    job_dict[wf][arg] = config.__dict__.get(arg)
                # raise error if still missing
                else:
                    raise KeyError(
                        f"Missing required argument '{arg}' for workflow '{wf}'"
                    )
    return job_dict


def launch_job(job_data: dict, terra: Terra, config: CLIConfig) -> None:
    """Launch a workflow in Terra based on provided job data"""
    current_time = datetime.now().strftime("%Y%m%d_%H%M%S")

    # could implement a way to use local table
    try:
        new_table_df = terra.entities.download_table(
            job_data["table"], use_destination=True
        )
    except TerraServerError as e:
        raise TerraServerError(
            f"Error downloading table; does \"{job_data['table']}\" exist in the Terra workspace?"
        )
    result = terra.entities.create_entity_set(
        f"{job_data['table']}_set_{current_time}", job_data["table"], new_table_df
    )
    if not result.ok:
        logger.error(f"Failed to create {job_data['table']} entity set")
        raise TerraServerError(f"Entity set creation failed: {result.text}")

    logger.info(f"{job_data['table']} entity set created successfully")

    # get method config dictionary from existing workspace
    if job_data.get("preexisting_config"):
        base_method_config_dict = terra.methods.get_method_config(
            job_data["workflow_name"], use_destination=True
        )
    else:
        base_method_config_dict = terra.methods.generate_method_config(
            config.repository,
            job_data["workflow_name"],
            job_data["table"],
            job_data["input_json"],
            job_data["output_json"],
            config.branch,
        )
    base_method_config = terra.methods.dict_to_method_config(base_method_config_dict)

    mod_method_config = base_method_config.model_copy(deep=True)
    # modify method config for the new workspace
    mod_method_config.namespace = terra.client.destination_project
    mod_method_config.rootEntityType = f"{job_data['table']}"
    mod_method_config.methodRepoMethod.methodUri = None
    mod_method_config.methodRepoMethod.sourceRepo = "dockstore"
    mod_method_config.methodRepoMethod.methodPath = (
        f"{config.repository}/{job_data['workflow_name']}"
    )
    mod_method_config.methodRepoMethod.methodVersion = config.branch
    mod_method_config.methodConfigVersion = 0

    # set inputs from json file and dynamically set samplename based on table name
    mod_method_config.inputs = job_data["input_json"]
    mod_method_config.inputs[f"{job_data['table']}.{job_data['sample_col']}"] = (
        f"this.{job_data['table']}_id"
    )
    # set outputs from json file
    mod_method_config.outputs = job_data["output_json"]

    # Adds or overwrites the method configuration in the Terra workspace
    terra.methods.overwrite_method_config(mod_method_config, use_destination=True)

    # Validate the new method configuration we created in the Terra workspace
    val = terra.methods.method_config_validate(mod_method_config, use_destination=True)
    # raise error if invalid I/O detected
    if val["invalidInputs"] or val["invalidOutputs"]:
        if val["invalidInputs"]:
            logger.error(f"Invalid inputs detected: {val['invalidInputs']}")
        if val["invalidOutputs"]:
            logger.error(f"Invalid outputs detected: {val['invalidOutputs']}")
        raise KeyError("Invalid I/O")

    wf_config_params = {
        "methodConfigurationNamespace": terra.client.destination_project,
        "methodConfigurationName": mod_method_config.name,
        "entityType": f"{job_data['table']}_set",  # entityType is name of set table
        "entityName": f"{job_data['table']}_set_{current_time}",  # entityName is name of specific row in table
        "expression": f"this.{job_data['table']}s",  # if rootEntityType is a set table, expression must be None. Otherwise, use this.{table_name}s format.
        "useCallCache": config.call_cache,
        "deleteIntermediateOutputFiles": False,
        "useReferenceDisks": False,
        "memoryRetryMultiplier": 1.0,
        "workflowFailureMode": "NoNewCalls",
        "userComment": job_data.get("comment", ""),
        "ignoreEmptyOutputs": config.ignore_empty,
    }

    # Workflow operations
    workflow_config = WorkflowConfig.model_validate(wf_config_params)
    submission = terra.submissions.submit_workflow(workflow_config)
    logger.info(submission)
    status = terra.submissions.get_submission_status(submission["submissionId"])
    logger.debug(status)


def launch(args: argparse.Namespace, config: CLIConfig = CLIConfig()) -> None:
    """Prepare and manage launch execution of workflows in Terra"""

    if args.workflow_name and args.job_json:
        raise ValueError("--workflow and --job_json are mutually exclusive")
    elif not args.workflow_name and not args.job_json:
        raise ValueError("--workflow or --job_json are required")
    elif args.workflow_name and (
        not args.table or not args.input_json or not args.output_json
    ):
        raise ValueError(
            "--table, --input_json, and --output_json are required when using --workflow"
        )

    config.update(vars(args))
    job_dicts = prepare_job_dict(vars(args), config)

    terra = Terra(
        source_workspace=config.workspace,
        source_project=config.project,
        destination_workspace=config.workspace,
        destination_project=config.project,
    )

    for wf_name, job_data in job_dicts.items():
        logger.info(f"Launching workflow: {wf_name}")
        launch_job(job_data, terra, config)
        time.sleep(args.sleep)



from .sample_processor import SampleDataProcessor
from .config_processor import ConfigProcessor
from .schema_models import (
    FieldAttributes,
    SampleFieldAttributes,
    ConfigFieldAttributes,
    FieldDefinition,
    SchemaDefinition,
)
from .schema_converter import (
    convert_field_attributes,
    convert_to_schema_definition,
    extract_field_attributes_dict,
)

__all__ = [
    "SampleDataProcessor",
    "ConfigProcessor",
    "FieldAttributes",
    "SampleFieldAttributes",
    "ConfigFieldAttributes",
    "FieldDefinition",
    "SchemaDefinition",
    "convert_field_attributes",
    "convert_to_schema_definition",
    "extract_field_attributes_dict",
]